#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 21 10:55:24 2019

Set of help functions for robust centring and scaling 

@author: Sven Serneels, Ponalytics
"""

import numpy as np
import pandas as ps
import scipy.stats as sps
import scipy.optimize as spo
from statsmodels import robust as srs
import copy

def _handle_zeros_in_scale(scale, copy=True):
    ''' 
    Makes sure that whenever scale is zero, we handle it correctly.
    This happens in most scalers when we have constant features.
    Taken from ScikitLearn.preprocesssing'''

    # if we are fitting on 1D arrays, scale might be a scalar
    if np.isscalar(scale):
        if scale == .0:
            scale = 1.
        return scale
    elif isinstance(scale, np.ndarray):
        if copy:
            # New array to avoid side-effects
            scale = scale.copy()
        scale[scale == 0.0] = 1.0
        return scale
    
def _check_trimming(t): 
    
    if ((t > .99) or (t < 0)): 
        raise(ValueError("Trimming fraction must be in [0,1)"))
    

def mad(X,c=0.6744897501960817,**kwargs):
        
    """
    Column-wise median absolute deviation. **kwargs included to allow 
    general function call in scale_data. 
    """
        
    s = np.median(np.abs(X - np.median(X,axis=0)),axis=0)/c
    s = np.array(s).reshape(-1)
        # statsmodels.robust.mad is not as flexible toward matrix input, 
        # sometimes throws a value error in ufunc
    return s

def median(X,**kwargs):
        
    """
    Column-wise median. **kwargs included to allow 
    general function call in scale_data. 
    """
        
    m = np.median(X,axis=0)
    m = np.array(m).reshape(-1)
    return m

def mean(X,trimming=0):
        
    """
    Column-wise mean or trimmed mean. Trimming to be entered as fraction. 
    """
        
    m = sps.trim_mean(X,trimming,0)
    return m

def std(X,trimming=0):
        
    """
    Column-wise standard devaition or trimmed std. 
    Trimming to be entered as fraction. 
    """
        
    if trimming==0:
        s = np.power(np.var(X,axis=0),.5)
        s = np.array(s).reshape(-1)
    else: 
        var = sps.trim_mean(np.square(X - sps.trim_mean(X,trimming,0)),
                            trimming,0)
        s = np.sqrt(var)   
    return s

def _euclidnorm(x):
    
        
    """
    Euclidean norm of a vector
    """
    
    return(np.sqrt(np.sum(np.square(x))))
    
def _diffmat_objective(a,X):
    
    """
    Utility to l1median, matrix of differences
    """
    
    (n,p) = X.shape
    return(X - np.tile(a,(n,1)))

def _l1m_objective(a,X,*args):
    
    """
    Optimization objective for l1median
    """
    
    return(np.sum(np.apply_along_axis(_euclidnorm,1,_diffmat_objective(a,X))))
    
def _l1m_jacobian(a,X):
    
    """
    Jacobian for l1median
    """
    
    (n,p) = X.shape
    dX = _diffmat_objective(a,X)
    dists = np.apply_along_axis(_euclidnorm,1,dX)
    dists = _handle_zeros_in_scale(dists)
    dX /= np.tile(np.array(dists).reshape(n,1),(1,p))
    return(-np.sum(dX,axis=0))

def _l1median(X,x0, method='SLSQP',tol=1e-8,options={'maxiter':2000},**kwargs): 
    
    """
    Optimization for l1median
    """
    
    mu = spo.minimize(_l1m_objective,x0,args=(X),
                      jac=_l1m_jacobian,tol=tol,options=options,method=method)
    return(mu)
    
    
def l1median(X,**kwargs): 
    
    """
    l1median wrapper to generically convert matrices as some of the scipy 
    optimization options will crash when provided matrix input. 
    """
    
    if 'x0' not in kwargs: 
        x0 = np.median(X,axis=0)
        x0 = np.array(x0.T).reshape(-1)

    if type(X) == np.matrix: 
        X = np.array(X)
        
    if len(X.shape) == 2:
        (n,p) = X.shape
    else:
        p = 1
    
    if p<2: 
        return(median(X))
    else:
        return(_l1median(X,x0,**kwargs).x)
        
def kstepLTS(X, maxit = 5, tol = 1e-10,**kwargs):
    
    """
    Computes the K-step LTS estimator of location
    It uses the spatial median as a starting value, and yields an 
    estimator with improved statistical efficiency, but at a higher 
    computational cost. 
    Inputs:
        X: data matrix
        maxit: maximum number of iterations
        tol: convergence tolerance
    Outputs:
        m2: location estimate
    """
    n,p = X.shape
    m1 = l1median(X) # initial estimate
    m2 = copy.deepcopy(m1)
    iteration = 0
    unconverged = True
    while(unconverged and (iteration < maxit)):
        dists = np.sum(np.square(X-m1),axis=1)
        cutdist = np.sort(dists,axis=0)[int(np.floor((n + 1) / 2))-1]
        hsubset = np.where(dists <= cutdist)[0]
        m2 = np.array(np.mean(X[hsubset,:],axis=0)).reshape((p,))
        unconverged = (max(abs(m1 - m2)) > tol)
        iteration += 1
        m1 = copy.deepcopy(m2)
    return(m2)
    
def scaleTau2(x0, c1 = 4.5, c2 = 3, consistency = True, **kwargs):   
    
    """
    Tau estimator of scale 
    Inputs:
        x0: array or matrix, data 
        c1: consistency factor for initial estimate 
        c2: consistency factor for final estimate 
        consistency: str or bool, 
            False, True, or "finiteSample" 
    Output:
        the scale estimate
    """
    
    x = copy.deepcopy(x0)
    n,p = x.shape
    medx = np.median(x,axis=0)
    xc = abs(x - medx)
    sigma0 = np.median(xc,axis=0) 
    if (c1 > 0):
        xc /= (sigma0 * c1)
        w = 1 - np.square(xc)
        w = np.square((abs(w) + w)/2)
        mu = np.sum(np.multiply(x,w))/np.sum(w)
    else: 
        mu = medx
    x -= mu
    x /= sigma0
    rho = np.square(x)
    rho[np.where(rho > c2**2)[0]] = c2**2
    if consistency:
        Erho = lambda b: 2 * ((1 - b**2) * sps.norm.cdf(b) - b * sps.norm.pdf(b) + 
            b**2) - 1
        Es2 = lambda c2: Erho(c2 * sps.norm.ppf(3/4))
        if (consistency == "finiteSample"):
            nEs2 = (n - 2) * Es2(c2)
        else: 
            nEs2 = n  * Es2(c2)
    else: 
        nEs2 = n
    return(np.array(sigma0 * np.sqrt(np.sum(rho)/nEs2)).reshape((p,)))
    
def scale_data(X,m,s):
        
        """
        Column-wise data scaling on location and scale estimates. 
        
        """
        
        n = X.shape
        if len(n) > 1:
            p = n[1]
        else:
            p = 1
        n = n[0]
        
        s = _handle_zeros_in_scale(s)
        
        if p == 1:
            Xm = X - float(m)
            Xs = Xm / s
        else:
            Xm = X - np.array([m for i in range(1,n+1)])
            Xs = Xm / np.array([s for i in range(1,n+1)])
        return(Xs)       

        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 27 11:41:35 2019

SPRM Plotting options (based on Matplotlib)
---------------------

In class sprm_plot
    Y vs Y predicted plot with outlier flagging
    Projection to score space with outlier flagging 
    Barchart of regression coefficients with range selection 
    Barchart of caseweights 
        All of these plots work both with or without new test datya beoing s
        supplied and have the optoion to only plot the test data.
        
In class sprm_plot_cv
    3D contour for SPRM Grid Search CV results 
        Requires an sklearn GridSearchCV object that trained an SPRM model 
        Provides a 3d contour plot across the (eta, n_components) space

Version 0.2: Ancillary functions: have been moved to ._plot_internals

@author: Sven Serneels, Ponalytics. 
"""

from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals

from ..sprm.sprm import sprm
from ..utils.utils import MyException, convert_X_input, convert_y_input
from ..cross_validation._cv_support_functions import cv_score_table

import matplotlib.pyplot as pp 
import numpy as np
from sklearn.model_selection import GridSearchCV

class sprm_plot(sprm):
    
    def __init__(self,res_sprm,colors,markers=['o','d','v'],*args):
        """
        Initialize with 
        res_sprm, an sprm class object
        
        Only mandatory input is colors, a list of colors for 
            [0] borders of pane 
            [1] plot background
            [2] marker fill
            [3] diagonal line 
            [4] marker contour, if different from fill
            [5] marker color for new cases, if applicable
            [6] marker color for harsh calibration outliers
            [7] marker color for harsh prediction outliers
        
        Optional input markers, a list:
            [0] marker for regular cases
            [1] marker for moderate outliers (caseweight in (0,1))
            [2] marker for harsh outliers (caseweight = 0)
        
        """
        if not(isinstance(res_sprm,sprm)):
            raise(MyException("Object supplied to sprmplot needs to be an sprm object"))
        self.res_sprm = res_sprm
        self.colors = colors
        self.markers = markers

    def plot_yyp(self,ytruev=[],Xn=[],label=[],names=[],namesv=[],title=[],legend_pos='lower right',onlyval=False):
        """
        plot_yyp will plot y vs y predicted for SPRM opbjects
        Optional inputs: 
            ytruev: array (new_cases,) of predictands
            Xn: array (new_cases,variables) of predictors 
            If these arguments are supplied, SPRM predictions for ytrue will be 
                made from Xn through res_sprm.predict()
            label: string: name of variable to be plotted. Will show in legend.
            names: list or tuple of strings, casenames from training set
            namesv: list or tuple of strings, casenames from test set
            title: String containing plot title
            legend_pos: string containing legend position
            onlyval: boolean: only plot validation cases
        """
        
        if len(label)==0:
            label = 'none'
        fig = pp.figure()
        fig.set_facecolor(self.colors[0])
        pp.rcParams['axes.facecolor'] = self.colors[1]
        ax1 = fig.add_subplot(111)
        if (not(onlyval)):
            ytruec = self.res_sprm.y
            if len(ytruec.shape) >1:
                ytruec = np.array(ytruec).reshape(-1).astype('float64')
            ypredc = np.array(self.res_sprm.fitted_).T.reshape(-1)
            labelcr = label[0] + ' Training' + ' Regular'
            labelcm = label[0] + ' Training' + ' Moderate'
            labelch = label[0] + ' Training' + ' Harsh'
            reg_cases = np.where(self.res_sprm.caseweights_ == 1)
            mod_outliers = np.where((self.res_sprm.caseweights_ > 0) & (self.res_sprm.caseweights_ < 1))
            harsh_outliers = np.where(self.res_sprm.caseweights_ == 0)
            ax1.scatter(ytruec[reg_cases], ypredc[reg_cases], c=self.colors[2],label=labelcr, 
                    zorder=1,edgecolors=self.colors[4],marker=self.markers[0])
            if len(mod_outliers[0]>0):
                ax1.scatter(ytruec[mod_outliers], ypredc[mod_outliers], c=self.colors[2],label=labelcm, 
                    zorder=1,edgecolors=self.colors[4],marker=self.markers[1])
            if len(harsh_outliers[0]>0):
                ax1.scatter(ytruec[harsh_outliers], ypredc[harsh_outliers], c=self.colors[6],label=labelch, 
                    zorder=1,edgecolors=self.colors[4],marker=self.markers[2])
        else:
            if (len(Xn)==0):
                ValueError('In onlyval=True mode, new cases Xn need to be provided')
        if not(len(Xn)==0):
            if len(ytruev.shape) >1:
                ytruev = np.array(ytruev).reshape(-1).astype('float64')
            ypredv = self.res_sprm.predict(Xn)
            ypredv = np.array(ypredv).reshape(-1).astype('float64')
            wv = self.res_sprm.weightnewx(Xn)
            labelvr = label[0] + ' Test' + ' Regular'
            labelvm = label[0] + ' Test' + ' Moderate'
            labelvh = label[0] + ' Test' + ' Harsh'
            reg_cases = np.where(wv == 1)[0]
            mod_outliers = np.where((wv > 0) & (wv < 1))[0]
            harsh_outliers = np.where(wv == 0)[0]
            if len(reg_cases>0):
                ax1.scatter(ytruev[reg_cases],ypredv[reg_cases],c=self.colors[5],label=labelvr,
                        zorder=1,edgecolors=self.colors[4],marker=self.markers[0])
            if len(mod_outliers>0):
                ax1.scatter(ytruev[mod_outliers],ypredv[mod_outliers],c=self.colors[5],label=labelvm,
                        zorder=1,edgecolors=self.colors[4],marker=self.markers[1])
            if len(harsh_outliers>0):
                ax1.scatter(ytruev[harsh_outliers],ypredv[harsh_outliers],c=self.colors[7],label=labelvh,
                        zorder=1,edgecolors=self.colors[4],marker=self.markers[2])
        x_abline = np.array(ax1.get_xbound())
        ax1.add_line(pp.Line2D(x_abline,x_abline,color=self.colors[3]))
        if len(label)==0:
            ax1.legend_.remove()
        else:
            pp.legend(loc=legend_pos)
        if len(names)>0:
            if not(onlyval):
                for i in range(0,len(names)-1):
                    ax1.annotate(names[i], (ytruec[i],ypredc[i]))
        if len(namesv)>0:
            for i in range(0,len(namesv)-1):
                ax1.annotate(namesv[i], (ytruev[i],ypredv[i]))
        if len(title)>0:
            pp.title(title)
        pp.show()
        
    def plot_coeffs(self,entity="coef_",truncation=0,columns=[],title=[]):
        """
        plot_coeffs will plot estimated model parameters, with option to 
        truncate (useful if highly multivariate) 
        Optional Inputs: 
            entity, str, exact name of attribute from sprm object to be plotted
            truncation, float in [0,1), Percentage of smallest and largest 
                coefficients to be plotted
            columns, array, tuple or list of str, variable names
            title, str, plot title
        """
        fig = pp.figure()
        fig.set_facecolor(self.colors[0])
        pp.rcParams['axes.facecolor'] = self.colors[1]
        ax1 = fig.add_subplot(111)
        p = len(self.res_sprm.non_zero_scale_vars_)
        if len(columns) > 0:
            x_plot = columns[self.res_sprm.non_zero_scale_vars_]
            x_labels = columns[self.res_sprm.non_zero_scale_vars_]
        else:
            x_plot = np.arange(0,p)
       
        b_orig = np.array(getattr(self.res_sprm,entity)).reshape(-1)
        if truncation > 0:
            ind_sort = np.argsort(np.argsort(b_orig))
            b_sort = np.sort(b_orig)
            left = np.ceil(p*(truncation/2)).astype(int)
            right = (p-np.ceil(p*(truncation/2)).astype(int))
            b_plot = b_sort[np.union1d(np.arange(0,left),
                                       np.arange(right,p))]
            x_labels = np.union1d(x_plot[ind_sort[0:left]],
                                x_plot[ind_sort[right:p]])
            x_plot = np.arange(0,len(x_labels))
            ax1.bar(x_plot,b_plot,color=self.colors[2],edgecolor=self.colors[4],tick_label=x_labels)
        else:
            b_plot = b_orig 
            ax1.bar(x_plot,b_plot,color=self.colors[2],edgecolor=self.colors[4])  
        if ((truncation > 0) | len(columns) > 0): 
            pp.xticks(x_plot, x_labels, rotation='vertical')
            pp.margins(.4)
        if len(title)>0:
            pp.title(title)
        pp.show()
        
        
    def plot_projections(self,Xn=[],label=[],components = [0,1],names=[],namesv=[],title=[],legend_pos='lower right',onlyval=False):
        
        """
        plot_projections will plot the score space  
        Optional inputs: 
            Xn: array (new_cases,variables) of predictors 
            If supplied, SPRM projections for new cases will be 
                made from Xn through res_sprm.transform()
            label: string: name of variable to be plotted. Will show in legend.
            names: list or tuple of strings, casenames from training set
            namesv: list or tuple of strings, casenames from test set
            title: String containing plot title
            legend_pos: string containing legend position
            onlyval: boolean: only plot validation cases
        """
        
        if len(label)==0:
            label = 'none'
        fig = pp.figure()
        fig.set_facecolor(self.colors[0])
        pp.rcParams['axes.facecolor'] = self.colors[1]
        ax1 = fig.add_subplot(111)
        if (not(onlyval)):
            Tc = np.array(self.res_sprm.x_scores_)
            labelcr = label[0] + ' Training' + ' Regular'
            labelcm = label[0] + ' Training' + ' Moderate'
            labelch = label[0] + ' Training' + ' Harsh'
            reg_cases = np.where(self.res_sprm.caseweights_ == 1)
            mod_outliers = np.where((self.res_sprm.caseweights_ > 0) & (self.res_sprm.caseweights_ < 1))
            harsh_outliers = np.where(self.res_sprm.caseweights_ == 0)
            ax1.scatter(Tc[reg_cases,components[0]], Tc[reg_cases,components[1]], c=self.colors[2],label=labelcr, 
                    zorder=1,edgecolors=self.colors[4],marker=self.markers[0])
            if len(mod_outliers[0])>0:
                ax1.scatter(Tc[mod_outliers,components[0]], Tc[mod_outliers,components[1]], c=self.colors[2],label=labelcm, 
                    zorder=1,edgecolors=self.colors[4],marker=self.markers[1])
            if len(harsh_outliers[0])>0:
                ax1.scatter(Tc[harsh_outliers,components[0]], Tc[harsh_outliers,components[1]], c=self.colors[6],label=labelch, 
                    zorder=1,edgecolors=self.colors[4],marker=self.markers[2])
        else:
            if (len(Xn)==0):
                ValueError('In onlyval=True mode, new cases Xn need to be provided')
        if not(len(Xn)==0):
            Tv = np.array(self.res_sprm.transform(Xn))
            labelvr = label[0] + ' Test' + ' Regular'
            labelvm = label[0] + ' Test' + ' Moderate'
            labelvh = label[0] + ' Test' + ' Harsh'
            wv = self.res_sprm.weightnewx(Xn)
            reg_cases = np.where(wv == 1)
            mod_outliers = np.where((wv > 0) & (wv < 1))
            harsh_outliers = np.where(wv == 0)
            if len(reg_cases[0]>0):
                ax1.scatter(Tv[reg_cases,components[0]], Tv[reg_cases,components[1]],c=self.colors[5],label=labelvr,
                        zorder=1,edgecolors=self.colors[4],marker=self.markers[0])
            if len(mod_outliers[0])>0:
                ax1.scatter(Tv[mod_outliers,components[0]], Tv[mod_outliers,components[1]],c=self.colors[5],label=labelvm,
                        zorder=1,edgecolors=self.colors[4],marker=self.markers[1])
            if len(harsh_outliers[0])>0:
                ax1.scatter(Tv[harsh_outliers,components[0]], Tv[harsh_outliers,components[1]],c=self.colors[7],label=labelvh,
                        zorder=1,edgecolors=self.colors[4],marker=self.markers[2])
        if len(label)==0:
            ax1.legend_.remove()
        else:
            pp.legend(loc=legend_pos)
        if len(names)>0:
            if not(onlyval):
                for i in range(0,len(names)-1):
                    ax1.annotate(names[i], (Tc[i,components[0]], Tc[i,components[1]]))
        if len(namesv)>0:
            for i in range(0,len(namesv)-1):
                ax1.annotate(namesv[i], (Tv[i,components[0]], Tv[i,components[1]]))
        if len(title)>0:
            pp.title(title)
        pp.show()
        
    def plot_caseweights(self,Xn=[],label=[],names=[],namesv=[],title=[],legend_pos='lower right',onlyval=False,mode='overall'):
        
        """
        plot_caseweights will plot caseweights
        Optional inputs: 
            Xn: array (new_cases,variables) of predictors 
            If supplied, SPRM projections for new cases will be 
                made from Xn through res_sprm.weightnewx()
            label: string: name of variable to be plotted. Will show in legend.
            names: list or tuple of strings, casenames from training set
            namesv: list or tuple of strings, casenames from test set
            title: String containing plot title
            legend_pos: string containing legend position
            onlyval: boolean: only plot validation cases
            mode: str, which weights to plot for cases from training set, 
                - 'overall': combined caseweights
                - 'x': predictor block 
                - 'y': predictand block
                Since for validation cases y is unknown, 'x' caseweights are 
                plotted by default there. 
        """
        
        if len(label)==0:
            label = 'none'
        fig = pp.figure()
        fig.set_facecolor(self.colors[0])
        pp.rcParams['axes.facecolor'] = self.colors[1]
        ax1 = fig.add_subplot(111)
        if (not(onlyval)):
            if mode=='overall':
                wc = self.res_sprm.caseweights_
            elif mode == 'x':
                wc = self.res_sprm.x_caseweights_
            elif mode == 'y':
                wc = self.res_sprm.y_caseweights_
            else:
                ValueError('Options for mode are overall, x or y')
            labelc = label[0] + ' Training' 
        else:
            wc = []
            labelc=[]
            if (len(Xn)==0):
                ValueError('In onlyval=True mode, new cases Xn need to be provided')
        if not(len(Xn)==0):
            wv = self.res_sprm.weightnewx(Xn)
            labelv = label[0] + ' Test'
        else:
            wv=[]
            labelv = []
        name_indices = np.array(range(1,len(wc)+len(wv)+1)) 
        if (len(wc)>0):
            ax1.bar(name_indices,np.concatenate((wc,np.repeat(np.nan,len(wv)))),color=self.colors[2],label=labelc)
        if (len(wv)>0):
            ax1.bar(name_indices,np.concatenate((np.repeat(np.nan,len(wc)),wv)),color=self.colors[5],label=labelv)
        if len(label)==0:
            ax1.legend_.remove()
        else:
            pp.legend(loc=legend_pos)
        if len(names)>0:
            if not(onlyval):
                for i in range(0,len(names)-1):
                    ax1.annotate(names[i], (name_indices[i+1],wc[i]))
        if len(namesv)>0:
            for i in range(0,len(namesv)-1):
                ax1.annotate(namesv[i], (name_indices[len(wc)+i],wv[i]))
        if len(title)>0:
            pp.title(title)
        pp.show()
        


class sprm_plot_cv(GridSearchCV,sprm):
    
    def __init__(self,res_sprm_cv,colors,*args):
        
        """
        Initialize with 
        res_sprm_cv, an GridSearchCV cross-validated sprm object 
        
        Only mandatory input is colors, a list of colors for 
            [0] borders of pane 
            [1] plot background
            [2] marker fill
            [3] diagonal line 
            [4] marker contour, if different from fill
            [5] marker color for new cases, if applicable
            [6] marker color for harsh calibration outliers
            [7] marker color for harsh prediction outliers
        
        """
        
        self.res_sprm_cv = res_sprm_cv
        self.colors = colors
        
    def eta_ncomp_contour(self,title='SPRM Cross-Validation Contour Plot'):
        
        """
        Function to draw contour plot from cross-valation results. 
        Optional Input: 
        title, str. Plot title. 
        
        """
        
        if not(hasattr(self,'cv_score_table_')):
            cv_score_table_ = cv_score_table(self.res_sprm_cv) 
            setattr(self,'cv_score_table_',cv_score_table_)
        fig = pp.figure()
        fig.set_facecolor(self.colors[0])
        pp.rcParams['axes.facecolor'] = self.colors[1]
        ax1 = fig.add_subplot(111)
        ax1.tricontour(self.cv_score_table_.values[:,0],self.cv_score_table_.values[:,1],self.cv_score_table_.values[:,2])
        pp.title(title)
        pp.show()          


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Fri Jan 26 13:56:12 2018

# Module containing:
    
#     Estimators
#     ----------
#     Sparse Partial Robust M Regression (SPRM)
    
#     0.2: Ancillary functions moved to ._m_support_functions
#          Plotting functions moved to .sprm_plot
#     0.3: Sparse NIPALS (SNIPLS) estimator moved to .snipls.

# Depends on robcent.VersatileScaler class for robustly centering and scaling data and on snipls 
# class

# @author: Sven Serneels, Ponalytics


from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals
from sklearn.base import RegressorMixin,BaseEstimator,TransformerMixin
from sklearn.utils.metaestimators import _BaseComposition
from scipy.stats import norm, chi2
import copy
import numpy as np
import pandas as ps
import warnings
from ..preprocessing.robcent import VersatileScaler
from .snipls import snipls
from ..utils.utils import MyException,_predict_check_input,_check_input
from ._m_support_functions import *
from ..preprocessing._preproc_utilities import scale_data, mad

class sprm(_BaseComposition,BaseEstimator,TransformerMixin,RegressorMixin):
    
    """
    SPRM Sparse Partial Robust M Regression 
    
    Algorithm first outlined in: 
        Sparse partial robust M regression, 
        Irene Hoffmann, Sven Serneels, Peter Filzmoser, Christophe Croux, 
        Chemometrics and Intelligent Laboratory Systems, 149 (2015), 50-59. 
    
    Parameters
    -----------

    eta : float.
          Sparsity parameter in [0,1)

    n_components : int
                     min 1. Note that if applied on data, n_components shall take a value <= min(x_data.shape)

    fun : str
         downweighting function. 'Hampel' (recommended), 'Fair' or  'Huber'

    probp1 : float
             probability cutoff for start of downweighting (e.g. 0.95)

    probp2 : float
            probability cutoff for start of steep downweighting (e.g. 0.975, only relevant if fun='Hampel')

    probp3 : float
            probability cutoff for start of outlier omission (e.g. 0.999, only relevant if fun='Hampel')

    centre : str
            type of centring (`'mean'`, `'median'`, `'l1median'`, or `'kstepLTS'`, 
            the latter recommended statistically, if too slow, switch to `'median'`)

    scale : str
            type of scaling ('std','mad', 'scaleTau2' [recommended] or 'None')

    verbose : booleans
             specifying verbose mode

    maxit : int
            maximal number of iterations in M algorithm

    tol : float
         tolerance for convergence in M algorithm 

    start_cutoff_mode : str,
                    values:'specific' will set starting value cutoffs specific to X and y (preferred); any other value will set X and y stating cutoffs identically. The latter yields identical results to the SPRM R implementation available from CRAN.
    start_X_init: str,
                 values: 'pcapp' will include a PCA/broken stick projection to  calculate the staring weights, else just based on X; any other value will calculate the X starting values based on the X matrix itself. This is less stable for very flat data (p >> n), yet yields identical results to the SPRM R implementation available from CRAN.   

    columns : (def false) Either boolean, list, numpy array or pandas Index
                if False, no column names supplied; if True, if X data are supplied as a pandas data frame, will extract column names from the frame throws an error for other data input types if a list, array or Index (will only take length x_data.shape[1]), the column names of the x_data supplied in this list, will be printed in verbose mode. 

    copy : (def True) boolean, whether to copy data

    Attributes
    ---------------
    Attributes always provided 

        -  `x_weights_`: X block PLS weighting vectors (usually denoted W)
        -  `x_loadings_`: X block PLS loading vectors (usually denoted P)
        -  `C_`: vector of inner relationship between response and latent variablesblock re
        -  `x_scores_`: X block PLS score vectors (usually denoted T)
        -  `coef_`: vector of regression coefficients 
        -  `intercept_`: intercept
        -  `coef_scaled_`: vector of scaled regression coeeficients (when scaling option used)
        -  `intercept_scaled_`: scaled intercept
        -  `residuals_`: vector of regression residuals
        -  `x_ev_`: X block explained variance per component
        -  `y_ev_`: y block explained variance 
        -  `fitted_`: fitted response
        -  `x_Rweights_`: X block SIMPLS style weighting vectors (usually denoted R)
        -  `x_caseweights_`: X block case weights
        -  `y_caseweights_`: y block case weights
        -  `caseweights_`: combined case weights
        -  `colret_`: names of variables retained in the sparse model
        -  `x_loc_`: X block location estimate 
        -  `y_loc_`: y location estimate
        -  `x_sca_`: X block scale estimate
        -  `y_sca_`: y scale estimate
        -  `non_zero_scale_vars_`: indicator vector of variables in X with nonzero scale

    
    """
    
    def __init__(self,n_components=1,eta=.5,fun='Hampel',probp1=0.95
                 ,probp2=0.975,probp3=0.999,centre='median',scale='mad'
                 ,verbose=True,maxit=100,tol=0.01,start_cutoff_mode='specific'
                 ,start_X_init='pcapp',columns=False,copy=True):
        self.n_components = int(n_components) 
        self.eta = float(eta)
        self.fun = fun
        self.probp1 = probp1
        self.probp2 = probp2
        self.probp3 = probp3
        self.centre = centre
        self.scale = scale
        self.verbose = verbose
        self.maxit = maxit
        self.tol = tol
        self.start_cutoff_mode = start_cutoff_mode
        self.start_X_init = start_X_init
        self.columns = columns
        self.copy = copy
        self.probctx_ = 'irrelevant'
        self.probcty_ = 'irrelevant'
        self.hampelbx_ = 'irrelevant'
        self.hampelby__ = 'irrelevant'
        self.hampelrx_ = 'irrelevant'
        self.hampelry_ = 'irrelevant'
        self.non_zero_scale_vars_ = None
    

    def fit(self,X,y):
        """
        Fit a  SPRM model. 
        
        Parameters
        ------------ 
            
            X : numpy array 
                Input data.

            y :   vector or 1D matrix
                Response data

        """
        if self.copy:
            self.X = copy.deepcopy(X)
            self.y = copy.deepcopy(y)
        (n,p) = X.shape
        if not(type(self.n_components)==int) | (self.n_components<=0):
            raise MyException("Number of components has to be a positive integer")
        if ((self.n_components > n) | (self.n_components > p)):
            raise MyException("The number of components is too large.")
        if (self.n_components <= 0):
            raise MyException("The number of components has to be positive.")
        if not(type(self.eta)==float):
            raise MyException("Sparsity parameter eta has to be a floating point number")
        if ((self.eta < 0) | (self.eta >= 1)):
            raise MyException("eta has to come from the interval [0,1)")
        if (not(self.fun in ("Hampel", "Huber", "Fair"))):
            raise MyException("Invalid weighting function. Choose Hampel, Huber or Fair for parameter fun.")
        if ((self.probp1 > 1) | (self.probp1 <= 0)):
            raise MyException("probp1 is a probability. Choose a value between 0 and 1")
        if (self.fun == "Hampel"):
            if (not((self.probp1 < self.probp2) & (self.probp2 < self.probp3) & (self.probp3 <= 1))):
                raise MyException("Wrong choise of parameters for Hampel function. Use 0<probp1<hampelp2<hampelp3<=1")
        if type(self.columns) is list: 
            self.columns = np.array(self.columns)
        elif type(self.columns) is bool:
            if type(X) != ps.core.frame.DataFrame and self.columns:
                raise(MyException("Columns set to true can only extract column names for data frame input"))
        if type(X) == ps.core.frame.DataFrame:
            if type(self.columns) is bool and self.columns:
                self.columns = X.columns
        X = _check_input(X)
        y = _check_input(y)
        ny = y.shape[0]
        if ny != n:
            raise MyException("Number of cases in y and X must be identical.")
            
        if self.scale == 'scaleTau2':
            if (mad(X)==0).any():
                # kstepLTS divides by MAD. Zero MAD leads to nan -> detect zero MAD, scale by MAD and remove zero MAD variable. 
                self.scale = 'mad'
                warnings.warn('Scale set to `mad`')

        scaling = VersatileScaler(center=self.centre, scale=self.scale)
        Xs = scaling.fit_transform(X).astype('float64')
        mX = scaling.col_loc_
        sX = scaling.col_sca_
        ys = scaling.fit_transform(y).astype('float64')
        my = scaling.col_loc_
        sy = scaling.col_sca_
        setattr(self,"x_loc_",mX)
        setattr(self,"y_loc_",my)
        setattr(self,"x_sca_",sX)
        setattr(self,"y_sca_",sy)
        ys = np.array(ys).reshape(-1)
        
        zero_scale = np.where(sX < 1e-5)[0]
        vars_to_keep = np.arange(0,p)
        if len(zero_scale) > 0:
            if type(self.columns) != bool:
                warntext = 'Zero scale variables with indices ' + str(self.columns[zero_scale]) + ' detected and removed'
            else:
                warntext = 'Zero scale variables with indices ' + str(zero_scale) + ' detected and removed'

            warnings.warn(warntext)
            vars_to_keep = np.setdiff1d(np.arange(0,p),zero_scale)
            Xs = Xs[:,vars_to_keep]
            X = X[:,vars_to_keep]
            sX = sX[vars_to_keep]
            if type(self.columns) != bool:
                self.columns = self.columns[vars_to_keep]
            p = len(vars_to_keep)
            if (self.n_components > p):
                raise MyException("Upon removal of zero scale variables, the number of components is too large. Please reduce to {p} maximally.")
        
        self.non_zero_scale_vars_ = vars_to_keep


        if (self.start_X_init=='pcapp'):
            U, S, V = np.linalg.svd(Xs)
            spc = np.square(S)
            spc /= np.sum(spc)
            relcomp = max(np.where(spc - brokenstick(min(p,n))[:,0] <=0)[0][0],1)
            Urc = np.array(U[:,0:relcomp])
            Us = scaling.fit_transform(Urc)
        else: 
            Us = Xs
        wx = np.sqrt(np.array(np.sum(np.square(Us),1),dtype=np.float64))
        wx = wx/np.median(wx)
        if [self.centre,self.scale]==['median','mad']:
            wy = np.array(abs(ys),dtype=np.float64)
        else:
            wy = (y - np.median(y,axis=0))/(1.4826*np.median(abs(y-np.median(y,axis=0)),axis=0))
        self.probcty_ = norm.ppf(self.probp1)
        if self.start_cutoff_mode == 'specific':
            self.probctx_ = chi2.ppf(self.probp1,relcomp)
        else: 
            self.probctx_ = self.probcty_
        if (self.fun == "Fair"):
            wx = Fair(wx,self.probctx_)
            wy = Fair(wy,self.probcty_)
        if (self.fun == "Huber"):
            wx = Huber(wx,self.probctx_)
            wy = Huber(wy,self.probcty_)
        if (self.fun == "Hampel"):
            self.hampelby_ = norm.ppf(self.probp2)
            self.hampelry_ = norm.ppf(self.probp3)
            if self.start_cutoff_mode == 'specific':
                self.hampelbx_ = chi2.ppf(self.probp2,relcomp)
                self.hampelrx_ = chi2.ppf(self.probp3,relcomp)
            else: 
                self.hampelbx_ = self.hampelby_
                self.hampelrx_ = self.hampelry_
            wx = Hampel(wx,self.probctx_,self.hampelbx_,self.hampelrx_)
            wy = Hampel(wy,self.probcty_,self.hampelby_,self.hampelry_)
        wx = np.array(wx).reshape(-1)
        wy = np.array(wy).reshape(-1)
        w = (wx*wy).astype("float64")
        if (w < 1e-06).any():
            w0 = np.where(w < 1e-06)[0]
            w[w0] = 1e-06
            we = np.array(w,dtype=np.float64)
        else:
            we = np.array(w,dtype=np.float64)
        wte = wx
        wye = wy
        WEmat = np.array([np.sqrt(we) for i in range(1,p+1)],ndmin=1).T   
        Xw = np.multiply(Xs,WEmat).astype("float64")
        yw = ys*np.sqrt(we)
        scalingt = copy.deepcopy(scaling)
        loops = 1
        rold = 1E-5
        difference = 1
        # Begin at iteration
        res_snipls = snipls(self.eta,self.n_components,
                            self.verbose,self.columns,'mean','None',
                            self.copy)
        while ((difference > self.tol) & (loops < self.maxit)):
            res_snipls.fit(Xw,yw)
            T = np.divide(res_snipls.x_scores_,WEmat[:,0:(self.n_components)])
            b = res_snipls.coef_
            yp = res_snipls.fitted_
            r = ys - yp
            if (len(r)/2 > np.sum(r == 0)):
                r = abs(r)/(1.4826 * np.median(abs(r)))
            else:
                r = abs(r)/(1.4826 * np.median(abs(r[r != 0])))
            scalet = self.scale
            if (scalet == "None"):
                scalingt.set_params(scale = "mad") 
            dt = scalingt.fit_transform(T)
            wtn = np.sqrt(np.array(np.sum(np.square(dt),1),dtype=np.float64))
            wtn = wtn/np.median(wtn)
            wtn = wtn.reshape(-1)
            wye = r.reshape(-1)
            wte = wtn
            if (self.fun == "Fair"):
                wte = Fair(wtn,self.probctx_)
                wye = Fair(wye,self.probcty_)
            if (self.fun == "Huber"):
                wte = Huber(wtn,self.probctx_)
                wye = Huber(wye,self.probcty_)
            if (self.fun == "Hampel"):
                self.probctx_ = chi2.ppf(self.probp1,self.n_components)
                self.hampelbx_ = chi2.ppf(self.probp2,self.n_components)
                self.hampelrx_ = chi2.ppf(self.probp3,self.n_components)
                wte = Hampel(wtn,self.probctx_,self.hampelbx_,self.hampelrx_)
                wye = Hampel(wye,self.probcty_,self.hampelby_,self.hampelry_)
            b2sum = np.sum(np.power(b,2))    
            difference = abs(b2sum - rold)/rold
            rold = b2sum
            wte = np.array(wte).reshape(-1)
            we = (wye * wte).astype("float64")
            w0=[]
            if (any(we < 1e-06)):
                w0 = np.where(we < 1e-06)[0]
                we[w0] = 1e-06
                we = np.array(we,dtype=np.float64)
            if (len(w0) >= (n/2)):
                break
            WEmat = np.array([np.sqrt(we) for i in range(1,p+1)],ndmin=1).T    
            Xw = np.multiply(Xs,WEmat).astype("float64")
            yw = ys*np.sqrt(we)
            loops += 1
        if (difference > self.maxit):
            print("Warning: Method did not converge. The scaled difference between norms of the coefficient vectors is " + 
                  str(round(difference,4)))
        plotprec = False
        if plotprec:
            print(str(loops - 1))
        w = we
        w[w0] = 0
        wt = wte
        wt[w0] = 0
        wy = wye
        wy[w0] = 0
        P = res_snipls.x_loadings_
        W = res_snipls.x_weights_
        R = res_snipls.x_Rweights_
        Xrw = np.array(np.multiply(Xs,np.sqrt(WEmat)).astype("float64"))
        scaling.set_params(scale='None')
        Xrw = scaling.fit_transform(Xrw) 
        T = np.matmul(Xs,R)
        if self.verbose:
            print("Final Model: Variables retained for " + str(self.n_components) + " latent variables: \n" 
                 + str(res_snipls.colret_) + "\n")
        b_rescaled = np.multiply(np.reshape(sy/sX,(p,1)),b)
        yp_rescaled = np.matmul(X,b_rescaled)
        if(self.centre == "mean"):
            intercept = np.mean(y - yp_rescaled,axis=0)
        else:
            intercept = np.median(y - yp_rescaled,axis=0)
        # This median calculation produces slightly different result in R and Py
        yfit = yp_rescaled + intercept   
        if (self.scale!="None"):
            if (self.centre == "mean"):
                b0 = np.mean(ys.astype("float64") - np.matmul(Xs.astype("float64"),b))
            else:
                b0 = np.median(np.array(ys.astype("float64") - np.matmul(Xs.astype("float64"),b)))
            # yfit2 = (np.matmul(Xrc.Xs.astype("float64"),b) + b0)*yrc.col_sca + yrc.col_loc
            # already more generally included
        else:
            if (self.centre == "mean"):
                b0 = np.mean(y - np.matmul(X,b))
            else:
                b0 = np.median(np.array(y - np.matmul(X,b)))
            # yfit = np.matmul(X,b) + intercept
        yfit = yfit
        r = y - yfit
        setattr(self,"x_weights_",W)
        setattr(self,"x_loadings_",P)
        setattr(self,"C_",res_snipls.C_)
        setattr(self,"x_scores_",T)
        setattr(self,"coef_",b_rescaled)
        setattr(self,"intercept_",intercept)
        setattr(self,"coef_scaled_",b)
        setattr(self,"intercept_scaled_",b0)
        setattr(self,"residuals_",r)
        setattr(self,"x_ev_",res_snipls.x_ev_)
        setattr(self,"y_ev_",res_snipls.y_ev_)
        setattr(self,"fitted_",yfit)
        setattr(self,"x_Rweights_",R)
        setattr(self,"x_caseweights_",wte)
        setattr(self,"y_caseweights_",wye)
        setattr(self,"caseweights_",we)
        setattr(self,"colret_",res_snipls.colret_)
        setattr(self,'scaling_',scaling)
        setattr(self,'scalingt_',scalingt)
        return(self)
        pass
    
        
    def predict(self,Xn):
        """
        Predict using a  SPRM model. 
        
        Parameters
        ------------ 
            
            Xn : numpy array or data frame 
                Input data.
        """
        n,p,Xn = _predict_check_input(Xn)
        if p!= self.X.shape[1]:
            raise(ValueError('New data must have seame number of columns as the ones the model has been trained with'))
        Xn = Xn[:,self.non_zero_scale_vars_]
        return(np.matmul(Xn,self.coef_) + self.intercept_)
        
    def transform(self,Xn):
        """
        Transform input data. 
        
        
        Parameters
        ------------ 
            
            Xn : numpy array or data frame 
                Input data.

        """
        n,p,Xn = _predict_check_input(Xn)
        if p!= self.X.shape[1]:
            raise(ValueError('New data must have seame number of columns as the ones the model has been trained with'))
        Xn = Xn[:,self.non_zero_scale_vars_]
        Xnc = scale_data(Xn,self.x_loc_[self.non_zero_scale_vars_],self.x_sca_[self.non_zero_scale_vars_])
        return(np.matmul(Xnc,self.x_Rweights_))
        
    def weightnewx(self,Xn):
        """
        Calculate case weights for new data based on the projection in the SPRM score space
        """
        n,p,Xn = _predict_check_input(Xn)
        (n,p) = Xn.shape
        if p!= self.X.shape[1]:
            raise(ValueError('New data must have seame number of columns as the ones the model has been trained with'))
        Tn = self.transform(Xn)
        scaling = self.scalingt_
        scalet = self.scale
        if (scalet == "None"):
            scaling.set_params(scale = "mad")
        if isinstance(Tn,np.matrix):
            Tn = np.array(Tn)
        dtn = scaling.fit_transform(Tn)
        wtn = np.sqrt(np.array(np.sum(np.square(dtn),1),dtype=np.float64))
        wtn = wtn/np.median(wtn)
        wtn = wtn.reshape(-1)
        if (self.fun == "Fair"):
            wtn = Fair(wtn,self.probctx_)
        if (self.fun == "Huber"):
            wtn = Huber(wtn,self.probctx_)
        if (self.fun == "Hampel"):
            wtn = Hampel(wtn,self.probctx_,self.hampelbx_,self.hampelrx_)
        return(wtn)
        
    def valscore(self,Xn,yn,scoring):
        """
        Specific score function for validation data
        """
        n,p,Xn = _predict_check_input(Xn)
        (n,p) = Xn.shape
        if p!= self.X.shape[1]:
            raise(ValueError('New data must have seame number of columns as the ones the model has been trained with'))
        if scoring=='weighted':
            return(RegressorMixin.score(self,Xn,yn,sample_weight=self.caseweights_))
        elif scoring=='normal':
            return(RegressorMixin.score(self,Xn,yn))
        else:
            raise(ValueError('Scoring flag must be set to "weighted" or "normal".'))

        
        
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Sun Dec 30 12:02:12 2018

# ppdire - Projection pursuit dimension reduction

# @author: Sven Serneels (Ponalytics)



#from .dicomo import dicomo
import numpy as np
from statsmodels.regression.quantile_regression import QuantReg
import statsmodels.robust as srs
import scipy.stats as sps
from scipy.linalg import pinv2
from scipy.optimize import minimize
import copy
from sklearn.utils.metaestimators import _BaseComposition
from sklearn.base import RegressorMixin,BaseEstimator,TransformerMixin, defaultdict
from sklearn.utils.extmath import svd_flip
from ..sprm.rm import rm 
from ..preprocessing.robcent import VersatileScaler
import warnings
from ..dicomo.dicomo import dicomo 
from ..dicomo._dicomo_utils import * 
from .capi import capi
from ._ppdire_utils import *
from ..preprocessing._preproc_utilities import scale_data
from ..utils.utils import MyException, convert_X_input, convert_y_input
import inspect

class ppdire(_BaseComposition,BaseEstimator,TransformerMixin,RegressorMixin):
    
    """
    PPDIRE Projection Pursuit Dimension Reduction
    
    The class allows for calculation of the projection pursuit optimization
    either through `scipy.optimize` or through the grid algorithm, native to this
    package. The class provides a very flexible way to access optimization of 
    projection indices that can lead to either classical or robust dimension
    reduction. Optimization through scipy.optimize is much more efficient, yet 
    it will only provide correct results for classical projection indices. The
    native grid algorithm should be used when the projection index involves 
    order statistics of any kind, such as ranks, trimming, winsorizing, or 
    empirical quantiles. The grid optimization algorithm for projection pursuit implemented here, 
    was outlined in: 
        
        Filzmoser, P., Serneels, S., Croux, C. and Van Espen, P.J., 
        Robust multivariate methods: The projection pursuit approach,
        in: From Data and Information Analysis to Knowledge Engineering,
        Spiliopoulou, M., Kruse, R., Borgelt, C., Nuernberger, A. and Gaul, W., eds., 
        Springer Verlag, Berlin, Germany,
        2006, pages 270--277.
        
    Parameters
    ------------ 

        projection_index : function or class. 
                            dicomo and capi supplied in this package can both be used, but user defined projection indices can be processed ball covariance can be used. 
            
        pi_arguments : dict 
                        arguments to be passed on to projection index 

        n_components : int 
                        number of components to estimate

        trimming : float
                     trimming percentage to be entered as pct/100 

        alpha : float.
                 Continuum coefficient. Only relevant if ppdire is used to estimate (classical or robust) continuum regression 

        optimizer : str.
                    Presently: either 'grid' (native optimizer) or any of the options in scipy-optimize (e.g. 'SLSQP')

        optimizer_options : dict 
                            with options to pass on to the optimizer 
                            If optimizer == 'grid',
                            ndir: int: Number of directions to calculate per iteration.
                            maxiter: int. Maximal number of iterations.

        optimizer_constraints : dict or list of dicts, 
                                further constraints to be passed on to the optimizer function.

        regopt : str. 
                regression option for regression step y~T. Can be set to 'OLS' (default), 'robust' (will run sprm.rm) or 'quantile' (statsmodels.regression.quantreg). 

        center : str, 
                how to center the data. options accepted are options from sprm.preprocessing 

        center_data : bool 

        scale_data : bool. 
                    Note: if set to False, convergence to correct optimum  is not a given. Will throw a warning. 

        whiten_data : bool. 
                    Typically used for ICA (kurtosis as PI)

        square_pi : bool. 
                    Whether to square the projection index upon evaluation.

        compression : bool. 
                        Use internal data compresion step for flat data. 

        copy : bool. 
                Whether to make a deep copy of the input data or not. 

        verbose : bool. 
                    Set to True prints the iteration number. 

        return_scaling_object : bool.
                                If True, the rescaling object will be returned. 

    Attributes
    ------------ 
    Attributes always provided 
        -  `x_weights_`: X block PPDIRE weighting vectors (usually denoted W)
        -  `x_loadings_`: X block PPDIRE loading vectors (usually denoted P)
        -  `x_scores_`: X block PPDIRE score vectors (usually denoted T)
        -  `x_ev_`: X block explained variance per component
        -  `x_Rweights_`: X block SIMPLS style weighting vectors (usually denoted R)
        -  `x_loc_`: X block location estimate 
        -  `x_sca_`: X block scale estimate
        -  `crit_values_`: vector of evaluated values for the optimization objective. 
        -  `Maxobjf_`: vector containing the optimized objective per component. 

    Attributes created when more than one block of data is provided: 
        -  `C_`: vector of inner relationship between response and latent variables block
        -  `coef_`: vector of regression coefficients, if second data block provided 
        -  `intercept_`: intercept
        -  `coef_scaled_`: vector of scaled regression coefficients (when scaling option used)
        -  `intercept_scaled_`: scaled intercept
        -  `residuals_`: vector of regression residuals
        -  `y_ev_`: y block explained variance 
        -  `fitted_`: fitted response
        -  `y_loc_`: y location estimate
        -  `y_sca_`: y scale estimate

    Attributes created only when corresponding input flags are `True`:
        -   `whitening_`: whitened data matrix (usually denoted K)
        -   `mixing_`: mixing matrix estimate
        -   `scaling_object_`: scaling object from `VersatileScaler`

    
    """

    def __init__(self,
                 projection_index, 
                 pi_arguments = {}, 
                 n_components = 1, 
                 trimming = 0,
                 alpha = 1,
                 optimizer = 'SLSQP',
                 optimizer_options = {'maxiter': 100000}, 
                 optimizer_constraints = {},
                 regopt = 'OLS',
                 center = 'mean',
                 center_data=True,
                 scale_data=True,
                 whiten_data=False,
                 square_pi = False,
                 compression = False,
                 copy=True,
                 verbose=True, 
                 return_scaling_object=True):
        # Called arguments
        self.projection_index = projection_index
        self.pi_arguments = pi_arguments
        self.n_components = n_components
        self.trimming = trimming
        self.alpha = alpha
        self.optimizer = optimizer
        self.optimizer_options = optimizer_options
        self.optimizer_constraints = optimizer_constraints
        self.regopt = regopt
        self.center = center
        self.center_data = center_data
        self.scale_data = scale_data
        self.whiten_data = whiten_data
        self.square_pi = square_pi
        self.compression = compression
        self.copy = copy
        self.verbose = verbose
        self.return_scaling_object = return_scaling_object
        
        # Other global parameters 
        self.constraint = 'norm'
        self.optrange = (-1,1)
        self.licenter = ['mean','median']
        if not(self.center in self.licenter):
            raise(ValueError('Only location estimator classes allowed are: "mean", "median"'))
    

    def fit(self,X,*args,**kwargs):
        
        """
            Fit a projection pursuit dimension reduction model. 

            Parameters
            ------------ 
                
                X : numpy array 
                    Input data.
        
        """

        # Collect optional fit arguments
        biascorr = kwargs.pop('biascorr',False)
            
        if 'h' not in kwargs:
            h = self.n_components
        else:
            h = kwargs.pop('h')
            self.n_components = h
            
        if 'dmetric' not in kwargs:
            dmetric = 'euclidean'
        else:
            dmetric = kwargs.get('dmetric')
            
        if 'mixing' not in kwargs:
            mixing = False
        else:
            mixing = kwargs.get('mixing')
            
        if 'y' not in kwargs:
            na = len(args)
            if na > 0: #Use of *args makes it sklearn consistent
                flag = 'two-block'
                y = args[0]
            else:
                flag = 'one-block'
                y = 0 # to allow calls with 'y=y' in spit of no real y argument present
        else:
            flag = 'two-block'
            y = kwargs.get('y')
                            
            if 'quantile' not in kwargs:
                quantile = .5
            else:
                quantile = kwargs.get('quantile')
                
            if self.regopt == 'robust':
            
                if 'fun' not in kwargs:
                    fun = 'Hampel'
                else:
                    fun = kwargs.get('fun')
                
                if 'probp1' not in kwargs:
                    probp1 = 0.95
                else:
                    probp1 = kwargs.get('probp1')
                
                if 'probp2' not in kwargs:
                    probp2 = 0.975
                else:
                    probp2 = kwargs.get('probp2')
                
                if 'probp3' not in kwargs:
                    probp3 = 0.99
                else:
                    probp3 = kwargs.get('probp3')

            
        if self.projection_index == dicomo:
            
            if self.pi_arguments['mode'] in ('M3','cos','cok'):
            
                if 'option' not in kwargs:
                    option = 1
                else:
                    option = kwargs.get('option')
                
                if option > 3:
                    print('Option value >3 will compute results, but meaning may be questionable')
                
        # Initiate projection index    
        self.most = self.projection_index(**self.pi_arguments)         
        
        # Initiate some parameters and data frames
        if self.copy:
            X0 = copy.deepcopy(X)
            self.X0 = X0
        else:
            X0 = X        
        X = convert_X_input(X0)    
        n,p = X0.shape 
        trimming = self.trimming
        
        # Check dimensions 
        if h > min(n,p):
            raise(MyException('number of components cannot exceed number of samples'))
            
        if (self.projection_index == dicomo and self.pi_arguments['mode'] == 'kurt' and self.whiten_data==False):
            warnings.warn('Whitening step is recommended for ICA')
            
        # Pre-processing adjustment if whitening
        if self.whiten_data:
            self.center_data = True
            self.scale_data = False
            self.compression = False
            print('All results produced are for whitened data')
        
        # Centring and scaling
        if self.scale_data:
            if self.center=='mean':
                scale = 'std'
            elif ((self.center=='median')|(self.center=='l1median')):
                scale = 'mad' 
        else:
            scale = 'None'
            warnings.warn('Without scaling, convergence to optima is not given')
            
         # Data Compression for flat tables if required                
        if ((p>n) and self.compression):
            V,S,U = np.linalg.svd(X.T,full_matrices=False)
            X = np.matmul(U.T,np.diag(S))
            n,p = X.shape
            
            if (srs.mad(X)==0).any(): 
                warnings.warn('Due to low scales in data, compression would induce zero scales.' 
                              + '\n' + 'Proceeding without compression.')
                dimensions = False
                if copy:
                    X = copy.deepcopy(X0)
                else:
                    X = X0
            else:
                dimensions = True
        else:
            dimensions = False
        
        # Initiate centring object and scale X data 
        centring = VersatileScaler(center=self.center,scale=scale,trimming=trimming)      
  
        if self.center_data:
            Xs = centring.fit_transform(X)
            mX = centring.col_loc_
            sX = centring.col_sca_
        else:
            Xs = X
            mX = np.zeros((1,p))
            sX = np.ones((1,p))

        fit_arguments = {}
            
        # Data whitening (best practice for ICA)
        if self.whiten_data:
            V,S,U = np.linalg.svd(Xs.T,full_matrices=False)
            del U
            K = (V/S)[:,:p]
            del V,S
            Xs = np.matmul(Xs, K)
            Xs *= np.sqrt(p)
        
        # Presently, X and y need to be matrices 
        # Will be changed to use regular np.ndarray
        Xs = np.matrix(Xs)

        # Pre-process y data when available 
        if flag != 'one-block':
            
            ny = y.shape[0]
            y = convert_y_input(y)
            if len(y.shape) < 2:
                y = np.matrix(y).reshape((ny,1))
#            py = y.shape[1]
            if ny != n:
                raise(MyException('X and y number of rows must agree'))
            if self.copy:
                y0 = copy.deepcopy(y)
                self.y0 = y0
                
            if self.center_data:
                ys = centring.fit_transform(y)
                my = centring.col_loc_
                sy = centring.col_sca_ 
            else:
                ys = y
                my = 0
                sy = 1
            ys = np.matrix(ys).astype('float64')
        
        else:
            ys = None
                

        # Initializing output matrices
        W = np.zeros((p,h))
        T = np.zeros((n,h))
        P = np.zeros((p,h))
        B = np.zeros((p,h))
        R = np.zeros((p,h))
        B_scaled = np.zeros((p,h))
        C = np.zeros((h,1))
        Xev = np.zeros((h,1))
        assovec = np.zeros((h,1))
        Maxobjf = np.zeros((h,1))

        # Initialize deflation matrices 
        E = copy.deepcopy(Xs)
        f = ys

        bi = np.zeros((p,1))
        
        opt_args = { 
                    'alpha': self.alpha,
                    'trimming': self.trimming,
                    'biascorr': biascorr, 
                    'dmetric' : 'euclidean',
                    }
        
        if self.optimizer=='grid':
            # Define grid optimization ranges
            if 'ndir' not in self.optimizer_options:
                self.optimizer_options['ndir'] = 1000
            optrange = np.sign(self.optrange)
            optmax = self.optrange[1]
            stop0s = np.arcsin(optrange[0])
            stop1s = np.arcsin(optrange[1])
            stop1c = np.arccos(optrange[0])
            stop0c = np.arccos(optrange[1])
            anglestart = max(stop0c,stop0s)
            anglestop = max(stop1c,stop1s)
            nangle = np.linspace(anglestart,anglestop,self.optimizer_options['ndir'],endpoint=False)            
            alphamat = np.matrix([np.cos(nangle), np.sin(nangle)])
            opt_args['_stop0c'] = stop0c
            opt_args['_stop0s'] = stop0s
            opt_args['_stop1c'] = stop1c
            opt_args['_stop1s'] = stop1s
            opt_args['optmax'] = optmax
            opt_args['optrange'] = self.optrange
            opt_args['square_pi'] = self.square_pi
            if optmax != 1:
                alphamat *= optmax
        
            if p>2:
                anglestart = min(opt_args['_stop0c'],opt_args['_stop0s'])
                anglestop = min(opt_args['_stop1c'],opt_args['_stop1s'])
                nangle = np.linspace(anglestart,anglestop,self.optimizer_options['ndir'],endpoint=True)
                alphamat2 = np.matrix([np.cos(nangle), np.sin(nangle)])
                if optmax != 1:
                    alphamat2 *= opt_args['optmax']
                
            # Arguments for grid plane
            opt_args['alphamat'] = alphamat,
            opt_args['ndir'] = self.optimizer_options['ndir'],
            opt_args['maxiter'] = self.optimizer_options['maxiter']
            if type(opt_args['ndir'] is tuple): 
                opt_args['ndir'] = opt_args['ndir'][0]
            
            # Arguments for grid plane #2
            grid_args_2 = { 
                     'alpha': self.alpha,
                     'alphamat': alphamat2,
                     'ndir': self.optimizer_options['ndir'],
                     'trimming': self.trimming,
                     'biascorr': biascorr, 
                     'dmetric' : 'euclidean',
                     '_stop0c' : stop0c,
                     '_stop0s' : stop0s,
                     '_stop1c' : stop1c,
                     '_stop1s' : stop1s,
                     'optmax' : optmax,
                     'optrange' : self.optrange,
                     'square_pi' : self.square_pi
                     }
            if flag=='two-block':
                grid_args_2['y'] = f
        
        if flag=='two-block':
            opt_args['y'] = f
            

        # Itertive coefficient estimation
        for i in range(0,h):

            if self.optimizer=='grid':
                if p==2:
                    wi,maximo = gridplane(E,self.most,
                                          pi_arguments=opt_args
                                          )
           
                elif p>2:
                
                    afin = np.zeros((p,1)) # final parameters for linear combinations
                    Z = copy.deepcopy(E)
                    # sort variables according to criterion
                    meas = [self.most.fit(E[:,k],
                            **opt_args) 
                            for k in np.arange(0,p)]
                    if self.square_pi:
                        meas = np.square(meas)
                    wi,maximo = gridplane(Z[:,0:2],self.most,opt_args)
                    Zopt = Z[:,0:2]*wi 
                    afin[0:2]=wi
                    for j in np.arange(2,p):
                        projmat = np.matrix([np.array(Zopt[:,0]).reshape(-1),
                                         np.array(Z[:,j]).reshape(-1)]).T
                        wi,maximo = gridplane(projmat,self.most,
                                              opt_args
                                              )
                        Zopt = Zopt*float(wi[0]) + Z[:,j]*float(wi[1])
                        afin[0:(j+1)] = afin[0:(j+1)]*float(wi[0])
                        afin[j] = float(wi[1])

                    tj = Z*afin
                    objf = self.most.fit(tj,
                                     **{**fit_arguments,**opt_args}
                                    )
                    if self.square_pi:
                        objf *= objf
    

                    # outer loop to run until convergence
                    objfold = copy.deepcopy(objf)
                    objf = -1000
                    afinbest = afin
                    ii = 0
                    maxiter_2j = 2**round(np.log2(self.optimizer_options['maxiter'])) 
                
                    while ((ii < self.optimizer_options['maxiter'] + 1) and (abs(objfold - objf)/abs(objf) > 1e-4)):
                        for j in np.arange(0,p):
                            projmat = np.matrix([np.array(Zopt[:,0]).reshape(-1),
                                         np.array(Z[:,j]).reshape(-1)]).T
                            if j > 16:
                                divv = maxiter_2j
                            else:
                                divv = min(2**j,maxiter_2j)
                        
                            wi,maximo = gridplane_2(projmat,
                                                    self.most,
                                                    q=afin[j],
                                                    div=divv,
                                                    pi_arguments=grid_args_2
                                                    )
                            Zopt = Zopt*float(wi[0,0]) + Z[:,j]*float(wi[1,0])
                            afin *= float(wi[0,0])
                            afin[j] += float(wi[1,0])
                        
                        # % evaluate the objective function:
                        tj = Z*afin
                    
                        objfold = copy.deepcopy(objf)
                        objf = self.most.fit(tj,
                                         q=afin,
                                         **opt_args
                                         )
                        if self.square_pi:
                            objf *= objf
                    
                        if  objf!=objfold:
                            if self.constraint == 'norm':
                                afinbest = afin/np.sqrt(np.sum(np.square(afin)))
                            else:
                                afinbest = afin
                            
                        ii +=1
                        if self.verbose:
                            print(str(ii))
                    #endwhile
                
                    afinbest = afin
                    wi = np.zeros((p,1))
                    wi = afinbest
                    Maxobjf[i] = objf
                # endif;%if p>2;
            else: # do not optimize by the grid algorithm
                if self.trimming > 0: 
                    warnings.warn('Optimization that involves a trimmed objective is not a quadratic program. The scipy-optimize result will be off!!')
                if 'center' in self.pi_arguments:
                    if (self.pi_arguments['center']=='median'): 
                        warnings.warn('Optimization that involves a median in the objective is not a quadratic program. The scipy-optimize result will be off!!')   
                constraint = {'type':'eq',
                              'fun': lambda x: np.linalg.norm(x) -1,
                              }
                if len(self.optimizer_constraints)>0: 
                    constraint = [constraint,self.optimizer_constraints]
                wi = minimize(pp_objective,
                              E[0,:].transpose(),
                              args=(self.most,E,opt_args),
                              method=self.optimizer,
                              constraints=constraint,
                              options=self.optimizer_options).x
                wi = np.matrix(wi).reshape((p,1))
                wi /= np.sqrt(np.sum(np.square(wi)))
                
                
            # Computing projection weights and scores
            ti = E*wi
            if self.optimizer != 'grid':
                Maxobjf[i] = self.most.fit(E*wi,**opt_args)
            nti = np.linalg.norm(ti)
            pi = E.T*ti / (nti**2)
            if self.whiten_data:
                wi /= np.sqrt((wi**2).sum())
                wi = K*wi
            wi0 = wi
            wi = np.array(wi)
            if len(W[:,i].shape) == 1:
                wi = wi.reshape(-1)
            W[:,i] = wi
            T[:,i] = np.array(ti).reshape(-1)
            P[:,i] = np.array(pi).reshape(-1)
            
            if flag != 'one-block':
                criteval = self.most.fit(E*wi0,
                                         **opt_args
                                         )
                if self.square_pi:
                    criteval *= criteval
                    
                assovec[i] = criteval
                

            # Deflation of the datamatrix guaranteeing orthogonality restrictions
            E -= ti*pi.T
 
            # Calculate R-Weights
            R = np.dot(W[:,0:(i+1)],pinv2(np.dot(P[:,0:(i+1)].T,W[:,0:(i+1)]),check_finite=False))
        
            # Execute regression y~T if y is present. Generate regression estimates.
            if flag != 'one-block':
                if self.regopt=='OLS':
                    ci = np.dot(ti.T,ys)/(nti**2)
                elif self.regopt == 'robust':
                    linfit = rm(fun=fun,probp1=probp1,probp2=probp2,probp3=probp3,
                                centre=self.center,scale=scale,
                                start_cutoff_mode='specific',verbose=self.verbose)
                    linfit.fit(ti,ys)
                    ci = linfit.coef_
                elif self.regopt == 'quantile':
                    linfit = QuantReg(y,ti)
                    model = linfit.fit(q=quantile)
                    ci = model.params
                # end regression if
                
                C[i] = ci
                bi = np.dot(R,C[0:(i+1)])
                bi_scaled = bi
                bi = np.multiply(np.reshape(sy/sX,(p,1)),bi)
                B[:,i] = bi[:,0]
                B_scaled[:,i] = bi_scaled[:,0]

        # endfor; Loop for latent dimensions

        # Re-adjust estimates to original dimensions if data have been compressed 
        if dimensions:
            B = np.matmul(V[:,0:p],B)
            B_scaled = np.matmul(V[:,0:p],B_scaled)
            R = np.matmul(V[:,0:p],R)
            W = np.matmul(V[:,0:p],W)
            P = np.matmul(V[:,0:p],P)
            bi = B[:,h-1]
            if self.center_data:
                Xs = centring.fit_transform(X0)
                mX = centring.col_loc_
                sX = centring.col_sca_
            else:
                Xs = X0
                mX = np.zeros((1,p))
                sX = np.ones((1,p))
        
        bi = bi.astype("float64")
        if flag != 'one-block':            
            # Calculate scaled and unscaled intercepts
            if dimensions:
                X = convert_X_input(X0)
            if(self.center == "mean"):
                intercept = sps.trim_mean(y - np.matmul(X,bi),trimming)
            else:
                intercept = np.median(np.reshape(y - np.matmul(X,bi),(-1)))
            yfit = np.matmul(X,bi) + intercept
            if not(scale == 'None'):
                if (self.center == "mean"):
                    b0 = np.mean(ys - np.matmul(Xs.astype("float64"),bi))
                else:
                    b0 = np.median(np.array(ys.astype("float64") - np.matmul(Xs.astype("float64"),bi)))
            else:
                b0 = intercept
            
            # Calculate fit values and residuals 
            yfit = yfit    
            r = y - yfit
            setattr(self,"coef_",B)
            setattr(self,"intercept_",intercept)
            setattr(self,"coef_scaled_",B_scaled)
            setattr(self,"intercept_scaled_",b0)
            setattr(self,"residuals_",r)
            setattr(self,"fitted_",yfit)
            setattr(self,"y_loadings_",C)
            setattr(self,"y_loc_",my)
            setattr(self,"y_sca_",sy)
                
        setattr(self,"x_weights_",W)
        setattr(self,"x_loadings_",P)
        setattr(self,"x_rotations_",R)
        setattr(self,"x_scores_",T)
        setattr(self,"x_ev_",Xev)
        setattr(self,"crit_values_",assovec)
        setattr(self,"Maxobjf_",Maxobjf)
        
        if self.whiten_data:
            setattr(self,"whitening_",K)

        
        if mixing:
            setattr(self,"mixing_",np.linalg.pinv(W))
        
        
        setattr(self,"x_loc_",mX)
        setattr(self,"x_sca_",sX)

        setattr(self,'scaling',scale)
        if self.return_scaling_object:
            setattr(self,'scaling_object_',centring)
        
        return(self)   


    def predict(self,Xn):
        """
        predicts the response  on new data Xn

        Parameters
        ----------
                Xn : matrix or data frame
                     Input data to be transformed 
                
        Returns
        -------
        predictions : numpy array 
                      The predictions from the dimension reduction model
        """
        Xn = convert_X_input(Xn)
        (n,p) = Xn.shape
        (q,h) = self.coef_.shape
        if p!=q:
            raise(ValueError('New data must have seame number of columns as the ones the model has been trained with'))
        return(np.array(np.matmul(Xn,self.coef_[:,h-1]) + self.intercept_).T.reshape(-1))
        
    def transform(self,Xn):
        """
        Computes the dimension reduction of the data Xn based on the fitted sudire model.

        Parameters
        ----------
                Xn : matrix or data frame
                     Input data to be transformed 

        Returns
        -------
        transformed_data : numpy array
                             the dimension reduced data 
        """
        Xn = convert_X_input(Xn)
        (n,p) = Xn.shape
        if p!= self.coef_.shape[0]:
            raise(ValueError('New data must have seame number of columns as the ones the model has been trained with'))
        Xnc = scale_data(Xn,self.x_loc_,self.x_sca_)
        return(Xnc*self.x_rotations_)
        
    @classmethod   
    def _get_param_names(cls):
        """Get parameter names for the estimator"""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []
    
        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError("scikit-learn estimators should always "
                                   "specify their parameters in the signature"
                                   " of their __init__ (no varargs)."
                                   " %s with constructor %s doesn't "
                                   " follow this convention."
                                   % (cls, init_signature))
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])
    
    def get_params(self, deep=False):
        """Get parameters for this estimator.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        ------
        Copied from ScikitLlearn instead of imported to avoid 'deep=True'
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key, None)
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out
        
    def set_params(self, **params):
        """Set the parameters of this estimator.
        Copied from ScikitLearn, adapted to avoid calling 'deep=True'
        Returns
        -------
        self
        ------
        Copied from ScikitLlearn instead of imported to avoid 'deep=True'
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params()
    
        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))
    
            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value
    
        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)
    
        return self
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul 22 12:18:53 2018

@author: Sven serneels, Ponalytics
"""

from setuptools import setup, find_packages
import re
import sys
import os

SRC_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)),"./src")
if SRC_DIR not in sys.path:
    sys.path.insert(0,SRC_DIR)
from direpack import __version__, __author__, __license__

readme_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'README.md')
try:
    from m2r import parse_from_file
    readme = parse_from_file(readme_file)
except ImportError:
    # m2r may not be installed in user environment
    with open(readme_file) as f:
        readme = f.read()

setup(
    name="direpack",
    version=__version__,
    author=__author__,
    author_email="svenserneels@gmail.com",
    description="A Python 3 Library for State-of-the-Art Statistical Dimension Reduction Methods",
    long_description=readme,
    long_description_content_type='text/markdown',
    url="https://github.com/SvenSerneels/direpack",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages('src'),  # include all packages under src
    package_dir={'':'src'},   # tell distutils packages are under src
    include_package_data = True,
    install_requires=[
        'numpy>=1.5.0',
        'scipy>=0.8.0',
        'matplotlib>=2.2.0',
        'scikit-learn>=0.18.0',
        'pandas>=0.19.0',
        'statsmodels>=0.8.0',
        # 'ipopt>=0.1.5',
        'dcor>=0.3'
    ]
)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 2 2020

@author: Sven Serneels, Ponalytics. 
"""

import numpy as np
import pandas as ps


def pp_objective(x,est,X,opt_args):
    
    """
    Optimization objective for ppdire 
    
    """
    
    n = len(x)
    x = np.matrix(x).reshape((n,1))
    return(-est.fit(np.matmul(X,x),**opt_args))

def gridplane(X,most,pi_arguments={},**kwargs):

    """
    Function for grid search in a plane in two dimensions
    
    Required: X, np.matrix(n,2), data, 
              most, class object, projection index. Designed for 
                  dicomo or capi classes.
    Optional: pi_arguments, dict: arguments to pass on to projection index, 
                    plus a few local arguments such as optrange and square_pi
                    (see ppdire for explanation)
        
              y, np.matrix(n,1), second block of data 
              biascorr, to apply bias correction at normal distribution 
              alphamat, np.matrix: matrix of alpha angles to be scanned. 
        
        
    Values: 
        wi, np.matrix(p,1): optimal direction 
        maximo, float: optimal value of projection index
        
    Note: this function is written exclusively to be called from within the ppdire class
    
    """
    
            
    if (('biascorr' not in kwargs) and ('biascorr' not in pi_arguments)):
        biascorr = False
    else:
        biascorr = kwargs.get('biascorr')
    
    if len(pi_arguments) == 0:
        
        pi_arguments = {
                        'alpha': 0,
                        'ndir': 1000,
                        'trimming': 0,
                        'biascorr': biascorr, 
                        'dmetric' : 'euclidean',
                        'alphamat': None,
                        'optrange': (-1,1),
                        'square_pi': False
                        }
        
    if ('y' in kwargs):
        y = kwargs.pop('y')
        pi_arguments['y'] = y
        
    optrange = pi_arguments['optrange']
    optmax = optrange[1]
    
    alphamat = kwargs.pop('alphamat',pi_arguments['alphamat'])
    if (alphamat != None):
        optrange = np.sign(optrange)
        stop0s = np.arcsin(optrange[0])
        stop1s = np.arcsin(optrange[1])
        stop1c = np.arccos(optrange[0])
        stop0c = np.arccos(optrange[1])
        anglestart = max(stop0c,stop0s)
        anglestop = max(stop1c,stop1s)
        nangle = np.linspace(anglestart,anglestop,pi_arguments['ndir'],endpoint=False)            
        alphamat = np.matrix([np.cos(nangle), np.sin(nangle)])
        if optmax != 1:
            alphamat *= optmax
    
    tj = X*alphamat
    if pi_arguments['square_pi']:
        meas = [most.fit(tj[:,i],**pi_arguments)**2 
        for i in np.arange(0,pi_arguments['ndir'])]
    else:
        meas = [most.fit(tj[:,i],**pi_arguments) 
        for i in np.arange(0,pi_arguments['ndir'])]
        
    maximo = np.max(meas)
    indmax = np.where(meas == maximo)[0]
    if len(indmax)>0:
        indmax = indmax[0]
    wi = alphamat[:,indmax]
    
    return(wi,maximo)
    
    

def gridplane_2(X,most,q,div,pi_arguments={},**kwargs):

    """
    Function for refining a grid search in a plane in two dimensions
    
    Required: X, np.matrix(n,2), data
              most, class object, projection index. Designed for 
                  dicomo or capi classes.
              q, np.matrix(1,1), last obtained suboptimal direction component
              div, float, number of subsegments to divide angle into
    
    Optional: pi_arguments, dict: arguments to pass on to projection index, 
                    plus a few local arguments such as optrange and square_pi
                    (see ppdire for explanation)
        
              y, np.matrix(n,1), second block of data 
              biascorr, to apply bias correction at normal distribution 
              alphamat, np.matrix: matrix of alpha angles to be scanned. 
        
    pi_arguments is a dict of arguments passed on to the projection index
        
    Values: 
        wi, np.matrix(p,1): optimal direction 
        maximo, float: optimal value of projection index
        
    Note: this function is written to be called from within the ppdire class
    
    """
            
    if (('biascorr' not in kwargs) and ('biascorr' not in pi_arguments)):
        biascorr = False
    else:
        biascorr = kwargs.get('biascorr')
        
    if len(pi_arguments) == 0:
        
        pi_arguments = {
                        'alpha': 0,
                        'ndir': 1000,
                        'trimming': 0,
                        'biascorr': biascorr, 
                        'dmetric' : 'euclidean',
                        'alphamat': None,
                        'optrange': (-1,1),
                        'square_pi': False
                        }

        
    if 'y' in kwargs:
        y = kwargs.pop('y')
        pi_arguments['y'] = y

    optrange = pi_arguments['optrange']
    optmax = optrange[1]
   
    alphamat = kwargs.pop('alphamat',pi_arguments['alphamat'])
    if (alphamat != None).any():
        anglestart = min(pi_arguments['_stop0c'],pi_arguments['_stop0s'])
        anglestop = min(pi_arguments['_stop1c'],pi_arguments['_stop1s'])
        nangle = np.linspace(anglestart,anglestop,pi_arguments['ndir'],endpoint=True)
        alphamat = np.matrix([np.cos(nangle), np.sin(nangle)])
        if optmax != 1:
            alphamat *= optmax
    alpha1 = alphamat
    divisor = np.sqrt(1 + 2*np.multiply(alphamat[0,:],alphamat[1,:])*q[0])
    alpha1 = np.divide(alphamat,np.repeat(divisor,2,0))
    tj = X*alpha1
    
    if pi_arguments['square_pi']:
        meas = [most.fit(tj[:,i],**pi_arguments)**2 
        for i in np.arange(0,pi_arguments['ndir'])]
    else:
        meas = [most.fit(tj[:,i],**pi_arguments) 
        for i in np.arange(0,pi_arguments['ndir'])]

    maximo = np.max(meas)
    indmax = np.where(meas == maximo)[0]
    if len(indmax)>0:
        indmax = indmax[0]
    wi = alpha1[:,indmax]
    
    return(wi,maximo)
